/* shapelib extension for PHP                                                *
 * Copyright 2004 Alexander Mayrhofer <axelm-php@nona.net>                   *
 * See COPYING for the status of this software                               *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify      *
 * it under the terms of the GNU General Public License as published by      *
 * the Free Software Foundation; either version 2 of the License, or         *
 * (at your option) any later version.                                       *
 *                                                                           *
 * This program is distributed in the hope that it will be useful,           * 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 * GNU General Public License for more details.	                             *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to the Free Software               *
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


/* include standard header */
#include "php.h"
#include "php_config.h"
/* configure output */
#include "config.h"

/* includes for shapelib */
#include <shapefil.h>

/* custom structs */
#include "ext_shapelib.h"

/* registered resource (SHPHandle) */
static int le_shphandle;
#define LE_SHPHANDLE_NAME "SHPHandle"

void shphandle_destr_handler(zend_rsrc_list_entry *rsrc TSRMLS_DC) {
	/* FIXME: Do whatever is neccessary to free shp/dbf */
	shpResource *res = (shpResource*)rsrc->ptr;
	dbf_field *flist, *next;

	if (res && res->hSHP) 
		SHPClose(res->hSHP); 
	if (res && res->hDBF) 
		DBFClose(res->hDBF);
	/* free field name list */
	flist = res->flist;
	while (flist) {
		if (flist->fname) free(flist->fname);
		next = flist->next;
		free(flist);
		flist = next;
	}
	res->flist = NULL;
	if (res) 
		free(res);
}

/* module initialization - register resource and constants */
PHP_MINIT_FUNCTION(shapelib) {
	/* Shape types */
	REGISTER_LONG_CONSTANT("SHPT_NULL", 0, 0);
	REGISTER_LONG_CONSTANT("SHPT_POINT", 1, 0);
	REGISTER_LONG_CONSTANT("SHPT_ARC", 3, 0);
	REGISTER_LONG_CONSTANT("SHPT_POLYGON", 5, 0);
	REGISTER_LONG_CONSTANT("SHPT_MULTIPOINT",	8, 0);
	REGISTER_LONG_CONSTANT("SHPT_POINTZ",	11, 0);
	REGISTER_LONG_CONSTANT("SHPT_ARCZ",	13, 0);
	REGISTER_LONG_CONSTANT("SHPT_POLYGONZ",	15, 0);
	REGISTER_LONG_CONSTANT("SHPT_MULTIPOINTZ", 18, 0);
	REGISTER_LONG_CONSTANT("SHPT_POINTM",	21, 0);
	REGISTER_LONG_CONSTANT("SHPT_ARCM",	23, 0);
	REGISTER_LONG_CONSTANT("SHPT_POLYGONM",	25, 0);
	REGISTER_LONG_CONSTANT("SHPT_MULTIPOINTM", 28, 0);
	REGISTER_LONG_CONSTANT("SHPT_MULTIPATCH", 31, 0);

	/* Shape primtives(?) */
	REGISTER_LONG_CONSTANT("SHPP_TRISTRIP",	0, 0);
	REGISTER_LONG_CONSTANT("SHPP_TRIFAN",	1, 0);
	REGISTER_LONG_CONSTANT("SHPP_OUTERRING",	2, 0);
	REGISTER_LONG_CONSTANT("SHPP_INNERRING",	3, 0);
	REGISTER_LONG_CONSTANT("SHPP_FIRSTRING",	4, 0);
	REGISTER_LONG_CONSTANT("SHPP_RING",	5, 0);

	/* dbf field types */
	REGISTER_LONG_CONSTANT("SHPFT_STRING",	FTString, 0);
	REGISTER_LONG_CONSTANT("SHPFT_INTEGER",	FTInteger, 0);
	REGISTER_LONG_CONSTANT("SHPFT_DOUBLE",	FTDouble, 0);
	REGISTER_LONG_CONSTANT("SHPFT_LOGICAL",	FTLogical, 0);
	REGISTER_LONG_CONSTANT("SHPFT_INVALID",	FTInvalid, 0);
	
	/* desctructors for out resource */
	le_shphandle = zend_register_list_destructors_ex
		(shphandle_destr_handler, NULL, LE_SHPHANDLE_NAME, 
		 module_number);
}

/* declaration of functions to be exported */
ZEND_FUNCTION(shp_create);	/* creates a shp-File, returns a handle */
ZEND_FUNCTION(shp_add);		/* adds a shape */
ZEND_FUNCTION(shp_close);	/* closes shape file */

/* compiled function list so Zend knows what's in this module */
zend_function_entry shapelib_functions[] =
{
    ZEND_FE(shp_create, NULL)
    ZEND_FE(shp_add,NULL)
    ZEND_FE(shp_close,NULL)
    {NULL, NULL, NULL}
};

/* compiled module information */
zend_module_entry shapelib_module_entry = {
    STANDARD_MODULE_HEADER,
    "shapelib module",      /* extension name */
    shapelib_functions,  /* extension function list */
    PHP_MINIT(shapelib), /* extension-wide startup function */
    NULL,               /* extension-wide shutdown function */
    NULL,               /* per-request startup function */
    NULL,               /* per-request shutdown function */
    NULL,               /* information function */
    "0.0.1",            /* extension version number (string) */
    STANDARD_MODULE_PROPERTIES
};

/* implement standard "stub" routine to introduce ourselves to Zend */
#if COMPILE_DL_EXT_SHAPELIB
ZEND_GET_MODULE(shapelib)
#endif

/* implementation of shp_create function */
ZEND_FUNCTION(shp_create)
{
	zval *shppath;
	zval *zval_handle;
	zval *shpfields;
	zval **shpfield;
	zval **val;
	long shptype;
	shpResource *shpres;
	SHPHandle hSHP;
	DBFHandle hDBF;
	int zr_handle; /* zend resource id for allocated handle */
	int fieldcount, i;
	long dbf_ftype;
	long dbf_flength;
	long dbf_fprecision;
	char *string_key; /* String key */  
	uint string_key_len; 
	dbf_field *cur;
	ulong num_key; /* Numeric key */
	HashPosition pos;
	dbf_field *flist, *last;

	if (ZEND_NUM_ARGS() != 3) {
		WRONG_PARAM_COUNT;
	}
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
	        "zlz", &shppath, &shptype, &shpfields) == FAILURE) {
	    WRONG_PARAM_COUNT; /* FIXME: not the correct error */
	}
	convert_to_string_ex(&shppath);

	if (shptype < SHPT_NULL || shptype > SHPT_MULTIPATCH) {
		zend_error(E_WARNING, "Illegal shapefile type %d", shptype);
		RETURN_NULL();
	}
	/* Trying to create the shp file(s) */
	if ((hSHP = SHPCreate(shppath->value.str.val, shptype)) && 
	    (hDBF = DBFCreate(shppath->value.str.val))) {
		/* allocating a zend resource for the handle */
		shpres = (shpResource*)malloc(sizeof(shpResource));
		/* FIXME: error when malloc fails */
		shpres->hSHP = hSHP;
		shpres->hDBF = hDBF;
		shpres->shp_type = shptype;
		shpres->flist = NULL;
		ZEND_REGISTER_RESOURCE(return_value, 
			(void*)shpres, le_shphandle);
		/* add specified fields to dbf */
		fieldcount = zend_hash_num_elements(Z_ARRVAL_PP(&shpfields));
		zend_hash_internal_pointer_reset_ex(Z_ARRVAL_PP(&shpfields), &pos);
		i=0;
		while (zend_hash_get_current_data_ex(Z_ARRVAL_PP(&shpfields), (void**)&shpfield, &pos) == SUCCESS) {

			if(Z_TYPE_PP(shpfield) != IS_ARRAY) {
				zend_error(E_WARNING, "dbf field %d - is of type %d", i, Z_TYPE_PP(shpfield));
				/* FIXME. warning above should be fatal */
			}
			if(zend_hash_get_current_key_ex(Z_ARRVAL_PP(&shpfields), &string_key, &string_key_len, &num_key, 1, &pos) == HASH_KEY_IS_LONG) {
				zend_error(E_WARNING, "dbf field names must be strings - longs are not supported");
				/* FIXME. warning above should be fatal */
			}
			// DEBUG zend_error(E_WARNING, "found field %s", string_key);
			/* fetching field type */
			if (zend_hash_index_find(Z_ARRVAL_PP(shpfield), 0, (void **)&val) == FAILURE) {
				zend_error(E_WARNING, "unable to fetch field typp for dbf field %s", string_key);
				/* FIXME, warning above should be fatal */
			}
			if (Z_TYPE_PP(val) != IS_LONG) {
				zend_error(E_WARNING, "field type for dbf field %s is not a long value", string_key);
				/* FIXME, warning above should be fatal */
			}
			if ((*val)->value.lval < FTString || (*val)->value.lval > FTInvalid) {
				zend_error(E_WARNING, "illegal field type %d for field %s", (*val)->value.lval, string_key);
				/* FIXME, warning above... */
			}
			dbf_ftype = (*val)->value.lval;
			/* fetching field length */
			if (zend_hash_index_find(Z_ARRVAL_PP(shpfield), 1, (void **)&val) == FAILURE) {
				zend_error(E_WARNING, "unable to fetch field typp for dbf field %s", string_key);
				/* FIXME, warning above should be fatal */
			}
			if (Z_TYPE_PP(val) != IS_LONG) {
				zend_error(E_WARNING, "field length for dbf field %s is not a long value", string_key);
				/* FIXME, warning above should be fatal */
			}
			dbf_flength = (*val)->value.lval;
			dbf_fprecision = 0;
			/* fetching field length */
			if (dbf_ftype == FTDouble) {
				if (zend_hash_index_find(Z_ARRVAL_PP(shpfield), 2, (void **)&val) == FAILURE) {
					zend_error(E_WARNING, "unable to fetch field typp for dbf field %s", string_key);
					/* FIXME, warning above should be fatal */
				}
				if (Z_TYPE_PP(val) != IS_LONG) {
					zend_error(E_WARNING, "field precision for dbf field %s is not a long value", string_key);
					/* FIXME, warning above should be fatal */
				}
			dbf_fprecision = (*val)->value.lval;
			}
			// DEBUG zend_error(E_WARNING, "adding field %s of type %d length %d, precision %d", string_key, dbf_ftype, dbf_flength, dbf_fprecision);
			/* append entry to field list */
			flist = (dbf_field*)malloc(sizeof(dbf_field));
			if (!flist) {
				zend_error(E_ERROR, "out of memory while allocating DFB field list");
				return;
			}
			/* FIXME: check of string_key < 10 (?) chars */
			/* FIXME: toupper() string_key ? */
			flist->fnum = DBFAddField(hDBF, string_key, dbf_ftype, dbf_flength, dbf_fprecision);
			flist->next = NULL;
			flist->fname = strdup(string_key);
			flist->ftype = dbf_ftype;
			flist->fsize = dbf_flength;
			flist->fprecision = dbf_fprecision;
			if (shpres->flist) {
				last->next = flist;
			} else {
				shpres->flist = flist;
			}
			last = flist;
			zend_hash_move_forward_ex(Z_ARRVAL_PP(&shpfields), &pos);
			i++;
		}
		return;
	}
	RETURN_NULL();
} /* end shp_create */

/* shp_add($shapehandle, $dbfarray(key=>val), array($x, $y, [$z]), arr.. ) */
ZEND_FUNCTION(shp_add)
{
	shpResource *shpres;
	SHPObject *obj;
	zval *arg_handle;
	zval *dbfvalues;
	zval **dbfvalue;
	zval *xyzarrays;
	zval **xyzarray;
	zval **val;
	char *string_key;
	uint string_key_len; 
	ulong num_key; 
	double x, y, z;
	double *xarr, *yarr, *zarr;
	int vertcount, numvertices, vertcur;
	int is_3d;
	int res;
	dbf_field *cur, *found;
	HashPosition pos;

	if (ZEND_NUM_ARGS() != 3) {
		WRONG_PARAM_COUNT;
	}

	/* fetch shape handle and dbf value array */
	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "raa", &arg_handle, &dbfvalues, &xyzarrays) == FAILURE) {
		    WRONG_PARAM_COUNT;
		    /* FIXME: this is probably the wrong error */
	}

	ZEND_FETCH_RESOURCE(shpres, shpResource*, &arg_handle, -1, LE_SHPHANDLE_NAME, le_shphandle); /* FIXME: error handling neccessary? */

	/* FIXME: can we trust zend_parse_parameters()? above */
	if(Z_TYPE_P(dbfvalues) != IS_ARRAY) {
		zend_error(E_ERROR, "expected hash as second parameter");
		return;
	}
	
	/* allocate xyz arrays to hand over to shapelib */
	numvertices = zend_hash_num_elements(Z_ARRVAL_P(xyzarrays));
	/* FIXME: does zend_error(E_ERROR ...) return? */
	if(!(xarr = (double*)malloc(numvertices*sizeof(double))))
		zend_error(E_ERROR, "failed to allocate vertices array");
	if(!(yarr = (double*)malloc(numvertices*sizeof(double))))
		zend_error(E_ERROR, "failed to allocate vertices array");
	if(!(zarr = (double*)malloc(numvertices*sizeof(double))))
		zend_error(E_ERROR, "failed to allocate vertices array");

	vertcount = 0;
	is_3d = 1;

	zend_hash_internal_pointer_reset_ex(Z_ARRVAL_PP(&xyzarrays), &pos);
	while(zend_hash_get_current_data_ex(Z_ARRVAL_PP(&xyzarrays), (void**)&xyzarray, &pos) == SUCCESS) {
		if(Z_TYPE_PP(xyzarray) != IS_ARRAY) {
			zend_error(E_WARNING, "third parameter: array member is not an array, skipping");
			zend_hash_move_forward_ex(Z_ARRVAL_PP(&xyzarrays), &pos);
			continue;
		}
		if(zend_hash_num_elements(Z_ARRVAL_PP(xyzarray)) < 2) {
			zend_error(E_WARNING, "third parameter: array member has less than 2 elements (x and y are required). skipping");
			zend_hash_move_forward_ex(Z_ARRVAL_PP(&xyzarrays), &pos);
			continue;
		}
		/* X */
		if(zend_hash_index_find(Z_ARRVAL_PP(xyzarray), 0, (void**)&val) == FAILURE) {
			zend_error(E_WARNING, "unable to fetch x coordinate, skipping");
			zend_hash_move_forward_ex(Z_ARRVAL_PP(&xyzarrays), &pos);
			continue;
		}
		convert_to_double_ex(val);
		xarr[vertcount] = (*val)->value.dval;
		/* Y */
		if(zend_hash_index_find(Z_ARRVAL_PP(xyzarray), 1, (void**)&val) == FAILURE) {
			zend_error(E_WARNING, "unable to fetch y coordinate, skipping");
			zend_hash_move_forward_ex(Z_ARRVAL_PP(&xyzarrays), &pos);
			continue;
		}
		convert_to_double_ex(val);
		yarr[vertcount] = (*val)->value.dval;
		/* Z - optional, but remember if any vertex was missing it */
		/* FIXME: does not work, but segfault when z not given */
		if(zend_hash_num_elements(Z_ARRVAL_PP(xyzarray)) > 2) {
		   	if(zend_hash_index_find(Z_ARRVAL_PP(xyzarray), 2, (void**)&val) == FAILURE) {
				is_3d = 0;
				vertcount++; /* Z optional, w're satisfied */
				zend_hash_move_forward_ex(Z_ARRVAL_PP(&xyzarrays), &pos);
				continue;
			} /* Z value found, add it */
			convert_to_double_ex(val);
			zarr[vertcount] = (*val)->value.dval;
		}
		vertcount++; 
		zend_hash_move_forward_ex(Z_ARRVAL_PP(&xyzarrays), &pos);
	}
	if (!is_3d) {
		free(zarr);
		zarr = NULL;
	}

	obj = SHPCreateSimpleObject(shpres->shp_type, vertcount, xarr, yarr, zarr);
	SHPRewindObject(shpres->hSHP, obj);
	res = SHPWriteObject(shpres->hSHP, -1, obj);
	SHPDestroyObject(obj);
	if (zarr) free(zarr);
	free(yarr);
	free(xarr);
	/* write dbf attributes */
	if(zend_hash_num_elements(Z_ARRVAL_P(dbfvalues)) < 1) {
		/* just write an empty DBF record */
		DBFWriteNULLAttribute(shpres->hDBF, res, 0);
		RETURN_LONG(res);
	}

	zend_hash_internal_pointer_reset_ex(Z_ARRVAL_PP(&dbfvalues), &pos);
	while (zend_hash_get_current_data_ex(Z_ARRVAL_PP(&dbfvalues), (void**)&dbfvalue, &pos) == SUCCESS) {
		if(zend_hash_get_current_key_ex(Z_ARRVAL_PP(dbfvalue), &string_key, &string_key_len, &num_key, 1, &pos) != HASH_KEY_IS_STRING) {
			zend_error(E_WARNING, "DBF field name is not a string");
			continue;
		}
		/* search for key in field list */
		cur = shpres->flist;
		found = NULL;
		while (cur && (!found)) {
			if (!strcasecmp(cur->fname, string_key)) {
				found = cur;
			}
			cur = cur->next;
		}
		if (found) {
			switch (found->ftype) {
			case FTString:
				convert_to_string_ex(dbfvalue);
				DBFWriteStringAttribute(shpres->hDBF, res, found->fnum, Z_STRVAL_PP(dbfvalue));
				break;
			case FTDouble:
				convert_to_double_ex(dbfvalue);
				DBFWriteDoubleAttribute(shpres->hDBF, res, found->fnum, Z_DVAL_PP(dbfvalue));
				break;
			case FTInteger:
				convert_to_long_ex(dbfvalue);
				DBFWriteIntegerAttribute(shpres->hDBF, res, found->fnum, Z_LVAL_PP(dbfvalue));
				break;
			default:
				zend_error(E_WARNING, "Unknown DBF field type %d", found->ftype);
			}
		} else {
			zend_error(E_WARNING, "DBF field '%s' not found, skipping", string_key);
		}
		zend_hash_move_forward_ex(Z_ARRVAL_PP(&dbfvalues), &pos);
	} /* end looping over dbfvalues */
	RETURN_LONG(res);
} /* end shp_add */ 
	
ZEND_FUNCTION(shp_close)
{
	shpResource *shpres;
	dbf_field *flist, *next;
	zval *arg_handle;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
	        "r", &arg_handle) == FAILURE) {
	    WRONG_PARAM_COUNT; /* FIXME: not the correct error */
	}
	
	ZEND_FETCH_RESOURCE(shpres, shpResource*, &arg_handle, -1, LE_SHPHANDLE_NAME, le_shphandle);
	
	/* close shape and dbf */
	SHPClose(shpres->hSHP); 
	DBFClose(shpres->hDBF);
	shpres->hSHP = NULL;
	shpres->hDBF = NULL;

	/* free field name list */
	flist = shpres->flist;
	while (flist) {
		if (flist->fname) free(flist->fname);
		next = flist->next;
		free(flist);
		flist = next;
	}
	shpres->flist = NULL;
	return; /* FIXME: set return values */
}
